@@ -24,10 +24,10 @@
 
 
 #include <gio/gunixinputstream.h>
-#include "flatpak-chain-input-stream.h"
+#include "flatpak-chain-input-stream-private.h"
 
 #include "flatpak-builtins-utils.h"
-#include "flatpak-utils.h"
+#include "flatpak-utils-private.h"
 
 
 gboolean
@@ -84,10 +84,11 @@ get_soup_session (void)
 }
 
 GBytes *
-download_uri (const char     *url,
-              GError        **error)
+download_uri (const char *url,
+              GError    **error)
 {
   SoupSession *session;
+
   g_autoptr(SoupRequest) req = NULL;
   g_autoptr(GInputStream) input = NULL;
   g_autoptr(GOutputStream) out = NULL;
@@ -121,6 +122,7 @@ flatpak_find_installed_pref (const char *pref, FlatpakKinds kinds, const char *d
   g_autofree char *id = NULL;
   g_autofree char *arch = NULL;
   g_autofree char *branch = NULL;
+
   g_autoptr(GError) lookup_error = NULL;
   FlatpakDir *dir = NULL;
   g_autofree char *ref = NULL;
@@ -255,7 +257,7 @@ flatpak_find_installed_pref (const char *pref, FlatpakKinds kinds, const char *d
 
 
 static gboolean
-open_source_stream (char **gpg_import,
+open_source_stream (char         **gpg_import,
                     GInputStream **out_source_stream,
                     GCancellable  *cancellable,
                     GError       **error)
@@ -304,7 +306,7 @@ open_source_stream (char **gpg_import,
 }
 
 GBytes *
-flatpak_load_gpg_keys (char **gpg_import,
+flatpak_load_gpg_keys (char        **gpg_import,
                        GCancellable *cancellable,
                        GError      **error)
 {
@@ -367,7 +369,7 @@ flatpak_resolve_duplicate_remotes (GPtrArray    *dirs,
         {
           FlatpakDir *dir = g_ptr_array_index (dirs_with_remote, i);
           g_autofree char *dir_name = flatpak_dir_get_name (dir);
-          g_print("%d) %s\n", i + 1, dir_name);
+          g_print ("%d) %s\n", i + 1, dir_name);
         }
       chosen = flatpak_number_prompt (0, dirs_with_remote->len, _("Which do you want to use (0 to abort)?"));
       if (chosen == 0)
@@ -377,11 +379,8 @@ flatpak_resolve_duplicate_remotes (GPtrArray    *dirs,
   if (out_dir)
     {
       if (dirs_with_remote->len == 0)
-        {
-          g_set_error (error, G_IO_ERROR, G_IO_ERROR_NOT_FOUND,
-                       "Remote \"%s\" not found", remote_name);
-          return FALSE;
-        }
+        return flatpak_fail_error (error, FLATPAK_ERROR_REMOTE_NOT_FOUND,
+                                   "Remote \"%s\" not found", remote_name);
       else
         *out_dir = g_object_ref (g_ptr_array_index (dirs_with_remote, chosen - 1));
     }
@@ -393,24 +392,25 @@ flatpak_resolve_duplicate_remotes (GPtrArray    *dirs,
 static guint64
 get_file_age (GFile *file)
 {
-    guint64 now;
-    guint64 mtime;
-    g_autoptr(GFileInfo) info = NULL;
+  guint64 now;
+  guint64 mtime;
 
-    info = g_file_query_info (file,
-                              G_FILE_ATTRIBUTE_TIME_MODIFIED,
-                              G_FILE_QUERY_INFO_NONE,
-                              NULL,
-                              NULL);
-    if (info == NULL)
-      return G_MAXUINT64;
+  g_autoptr(GFileInfo) info = NULL;
 
-    mtime = g_file_info_get_attribute_uint64 (info, G_FILE_ATTRIBUTE_TIME_MODIFIED);
-    now = (guint64) g_get_real_time () / G_USEC_PER_SEC;
-    if (mtime > now)
-      return G_MAXUINT64;
+  info = g_file_query_info (file,
+                            G_FILE_ATTRIBUTE_TIME_MODIFIED,
+                            G_FILE_QUERY_INFO_NONE,
+                            NULL,
+                            NULL);
+  if (info == NULL)
+    return G_MAXUINT64;
 
-    return (guint64) (now - mtime);
+  mtime = g_file_info_get_attribute_uint64 (info, G_FILE_ATTRIBUTE_TIME_MODIFIED);
+  now = (guint64) g_get_real_time () / G_USEC_PER_SEC;
+  if (mtime > now)
+    return G_MAXUINT64;
+
+  return (guint64) (now - mtime);
 }
 
 static void
@@ -418,6 +418,22 @@ no_progress_cb (OstreeAsyncProgress *progress, gpointer user_data)
 {
 }
 
+static guint64
+get_appstream_timestamp (FlatpakDir *dir,
+                         const char *remote,
+                         const char *arch)
+{
+  g_autoptr(GFile) ts_file = NULL;
+  g_autofree char *ts_file_path = NULL;
+  g_autofree char *subdir = NULL;
+
+  subdir = g_strdup_printf ("appstream/%s/%s/.timestamp", remote, arch);
+  ts_file = g_file_resolve_relative_path (flatpak_dir_get_path (dir), subdir);
+  ts_file_path = g_file_get_path (ts_file);
+  return get_file_age (ts_file);
+}
+
+
 gboolean
 update_appstream (GPtrArray    *dirs,
                   const char   *remote,
@@ -432,15 +448,16 @@ update_appstream (GPtrArray    *dirs,
   int i, j;
 
   g_return_val_if_fail (dirs != NULL, FALSE);
-  g_return_val_if_fail (arch != NULL, FALSE);
+
+  if (arch == NULL)
+    arch = flatpak_get_arch ();
 
   if (remote == NULL)
     {
-      g_auto(GStrv) remotes = NULL;
-
       for (j = 0; j < dirs->len; j++)
         {
           FlatpakDir *dir = g_ptr_array_index (dirs, j);
+          g_auto(GStrv) remotes = NULL;
 
           remotes = flatpak_dir_list_remotes (dir, cancellable, error);
           if (remotes == NULL)
@@ -450,47 +467,50 @@ update_appstream (GPtrArray    *dirs,
             {
               g_autoptr(GError) local_error = NULL;
               g_autoptr(OstreeAsyncProgress) progress = NULL;
-              g_autoptr(GFile) ts_file = NULL;
-              g_autofree char *ts_file_path = NULL;
-              g_autofree char *subdir = NULL;
               guint64 ts_file_age;
 
-              subdir = g_strdup_printf ("appstream/%s/%s/.timestamp", remotes[i], arch);
-              ts_file = g_file_resolve_relative_path (flatpak_dir_get_path (dir), subdir);
-              ts_file_path = g_file_get_path (ts_file);
-              ts_file_age = get_file_age (ts_file);
+              ts_file_age = get_appstream_timestamp (dir, remotes[i], arch);
               if (ts_file_age < ttl)
                 {
-                  g_debug ("%s age %" G_GUINT64_FORMAT " is less than ttl %" G_GUINT64_FORMAT, ts_file_path, ts_file_age, ttl);
+                  g_debug ("%s:%s appstream age %" G_GUINT64_FORMAT " is less than ttl %" G_GUINT64_FORMAT, remotes[i], arch, ts_file_age, ttl);
                   continue;
                 }
               else
-                g_debug ("%s age %" G_GUINT64_FORMAT " is greater than ttl %" G_GUINT64_FORMAT, ts_file_path, ts_file_age, ttl);
-
+                g_debug ("%s:%s appstream age %" G_GUINT64_FORMAT " is greater than ttl %" G_GUINT64_FORMAT, remotes[i], arch, ts_file_age, ttl);
 
               if (flatpak_dir_get_remote_disabled (dir, remotes[i]) ||
-                  flatpak_dir_get_remote_noenumerate (dir, remotes[i]) ||
-                  !flatpak_dir_check_for_appstream_update (dir, remotes[i], arch))
+                  flatpak_dir_get_remote_noenumerate (dir, remotes[i]))
                 continue;
 
               if (flatpak_dir_is_user (dir))
                 {
                   if (quiet)
-                    g_debug (_("Updating appstream data for user remote %s\n"), remotes[i]);
+                    g_debug (_("Updating appstream data for user remote %s"), remotes[i]);
                   else
-                    g_print (_("Updating appstream data for user remote %s\n"), remotes[i]);
+                    {
+                      g_print (_("Updating appstream data for user remote %s"), remotes[i]);
+                      g_print ("\n");
+                    }
                 }
               else
                 {
                   if (quiet)
-                    g_debug (_("Updating appstream data for remote %s\n"), remotes[i]);
+                    g_debug (_("Updating appstream data for remote %s"), remotes[i]);
                   else
-                    g_print (_("Updating appstream data for remote %s\n"), remotes[i]);
+                    {
+                      g_print (_("Updating appstream data for remote %s"), remotes[i]);
+                      g_print ("\n");
+                    }
                 }
               progress = ostree_async_progress_new_and_connect (no_progress_cb, NULL);
               if (!flatpak_dir_update_appstream (dir, remotes[i], arch, &changed,
                                                  progress, cancellable, &local_error))
-                g_printerr (_("Error updating: %s\n"), local_error->message);
+                {
+                  if (quiet)
+                    g_debug ("%s: %s", _("Error updating"), local_error->message);
+                  else
+                    g_printerr ("%s: %s\n", _("Error updating"), local_error->message);
+                }
               ostree_async_progress_finish (progress);
             }
         }
@@ -503,15 +523,21 @@ update_appstream (GPtrArray    *dirs,
         {
           FlatpakDir *dir = g_ptr_array_index (dirs, j);
 
-          if (flatpak_dir_has_remote (dir, remote))
+          if (flatpak_dir_has_remote (dir, remote, NULL))
             {
               g_autoptr(OstreeAsyncProgress) progress = NULL;
+              guint64 ts_file_age;
 
               found = TRUE;
 
-              /* Early bail out check */
-              if (!flatpak_dir_check_for_appstream_update (dir, remote, arch))
-                continue;
+              ts_file_age = get_appstream_timestamp (dir, remote, arch);
+              if (ts_file_age < ttl)
+                {
+                  g_debug ("%s:%s appstream age %" G_GUINT64_FORMAT " is less than ttl %" G_GUINT64_FORMAT, remote, arch, ts_file_age, ttl);
+                  continue;
+                }
+              else
+                g_debug ("%s:%s appstream age %" G_GUINT64_FORMAT " is greater than ttl %" G_GUINT64_FORMAT, remote, arch, ts_file_age, ttl);
 
               progress = ostree_async_progress_new_and_connect (no_progress_cb, NULL);
               res = flatpak_dir_update_appstream (dir, remote, arch, &changed,
@@ -523,7 +549,8 @@ update_appstream (GPtrArray    *dirs,
         }
 
       if (!found)
-        return flatpak_fail (error, _("Remote \"%s\" not found"), remote);
+        return flatpak_fail_error (error, FLATPAK_ERROR_REMOTE_NOT_FOUND,
+                                   _("Remote \"%s\" not found"), remote);
     }
 
   return TRUE;
