@@ -5,6 +5,7 @@
 #include <sys/wait.h>
 
 #include <glib.h>
+#include <ostree.h>
 
 #include "libglnx/libglnx.h"
 #include "flatpak.h"
@@ -18,19 +19,26 @@ static char *flatpak_installationsdir;
 static char *gpg_homedir;
 static char *gpg_args;
 static char *repo_url;
-#ifdef FLATPAK_ENABLE_P2P
 static char *repo_collection_id;
-#endif  /* FLATPAK_ENABLE_P2P */
 int httpd_pid = -1;
 
 static const char *gpg_id = "7B0961FD";
 const char *repo_name = "test-repo";
 
+typedef enum {
+  RUN_TEST_SUBPROCESS_DEFAULT = 0,
+  RUN_TEST_SUBPROCESS_IGNORE_FAILURE = (1 << 0),
+  RUN_TEST_SUBPROCESS_NO_CAPTURE = (1 << 1),
+} RunTestSubprocessFlags;
+
+static void run_test_subprocess (char                 **argv,
+                                 RunTestSubprocessFlags flags);
+
 typedef struct
 {
-  const char *id;
-  const char *display_name;
-  gint        priority;
+  const char        *id;
+  const char        *display_name;
+  gint               priority;
   FlatpakStorageType storage_type;
 } InstallationExtraData;
 
@@ -39,7 +47,7 @@ test_library_version (void)
 {
   g_autofree char *version = NULL;
 
-  version = g_strdup_printf ("%d.%d.%d" G_STRINGIFY(PACKAGE_EXTRA_VERSION),
+  version = g_strdup_printf ("%d.%d.%d" G_STRINGIFY (PACKAGE_EXTRA_VERSION),
                              FLATPAK_MAJOR_VERSION,
                              FLATPAK_MINOR_VERSION,
                              FLATPAK_MICRO_VERSION);
@@ -114,11 +122,11 @@ test_multiple_system_installations (void)
 
   for (i = 0; i < system_dirs->len; i++)
     {
-      g_autoptr (FlatpakInstallation) new_install = NULL;
+      g_autoptr(FlatpakInstallation) new_install = NULL;
       g_autoptr(GFile) installation_path = NULL;
       g_autofree char *path_str = NULL;
 
-      installation = (FlatpakInstallation*) g_ptr_array_index (system_dirs, i);
+      installation = (FlatpakInstallation *) g_ptr_array_index (system_dirs, i);
       g_assert_false (flatpak_installation_get_is_user (installation));
 
       installation_path = flatpak_installation_get_path (installation);
@@ -147,10 +155,10 @@ test_multiple_system_installations (void)
       new_install = flatpak_installation_new_system_with_id (current_id, NULL, &error);
       g_assert_nonnull (new_install);
 
-      g_assert_cmpstr(current_id, ==, flatpak_installation_get_id (new_install));
-      g_assert_cmpstr(current_display_name, ==, flatpak_installation_get_display_name (new_install));
-      g_assert_cmpint(current_priority, ==, flatpak_installation_get_priority (new_install));
-      g_assert_cmpint(current_storage_type, ==, flatpak_installation_get_storage_type (new_install));
+      g_assert_cmpstr (current_id, ==, flatpak_installation_get_id (new_install));
+      g_assert_cmpstr (current_display_name, ==, flatpak_installation_get_display_name (new_install));
+      g_assert_cmpint (current_priority, ==, flatpak_installation_get_priority (new_install));
+      g_assert_cmpint (current_storage_type, ==, flatpak_installation_get_storage_type (new_install));
     }
 }
 
@@ -211,9 +219,7 @@ test_ref (void)
   g_assert_cmpstr (flatpak_ref_get_name (ref), ==, "org.flatpak.Hello");
   g_assert_cmpstr (flatpak_ref_get_arch (ref), ==, "x86_64");
   g_assert_cmpstr (flatpak_ref_get_branch (ref), ==, "master");
-#ifdef FLATPAK_ENABLE_P2P
   g_assert_null (flatpak_ref_get_collection_id (ref));
-#endif  /* FLATPAK_ENABLE_P2P */
 
   formatted = flatpak_ref_format_ref (ref);
   g_assert_cmpstr (formatted, ==, valid);
@@ -225,7 +231,10 @@ test_list_remotes (void)
   g_autoptr(FlatpakInstallation) inst = NULL;
   g_autoptr(GError) error = NULL;
   g_autoptr(GPtrArray) remotes = NULL;
+  g_autoptr(GPtrArray) remotes2 = NULL;
   FlatpakRemote *remote;
+  const FlatpakRemoteType types[] = { FLATPAK_REMOTE_TYPE_STATIC };
+  const FlatpakRemoteType types2[] = { FLATPAK_REMOTE_TYPE_LAN };
 
   inst = flatpak_installation_new_user (NULL, &error);
   g_assert_no_error (error);
@@ -237,6 +246,30 @@ test_list_remotes (void)
 
   remote = g_ptr_array_index (remotes, 0);
   g_assert (FLATPAK_IS_REMOTE (remote));
+
+  remotes2 = flatpak_installation_list_remotes_by_type (inst, types,
+                                                        G_N_ELEMENTS (types),
+                                                        NULL, &error);
+  g_assert_no_error (error);
+  g_assert_cmpuint (remotes2->len, ==, remotes->len);
+
+  for (guint i = 0; i < remotes->len; ++i)
+    {
+      FlatpakRemote *remote1 = g_ptr_array_index (remotes, i);
+      FlatpakRemote *remote2 = g_ptr_array_index (remotes2, i);
+      g_assert_cmpstr (flatpak_remote_get_name (remote1), ==,
+                       flatpak_remote_get_name (remote2));
+      g_assert_cmpstr (flatpak_remote_get_url (remote1), ==,
+                       flatpak_remote_get_url (remote2));
+    }
+
+  g_ptr_array_unref (remotes2);
+  remotes2 = flatpak_installation_list_remotes_by_type (inst,
+                                                        types2,
+                                                        G_N_ELEMENTS (types2),
+                                                        NULL, &error);
+  g_assert_no_error (error);
+  g_assert_cmpuint (remotes2->len, ==, 0);
 }
 
 static void
@@ -262,9 +295,7 @@ test_remote_by_name (void)
   g_assert_true (flatpak_remote_get_gpg_verify (remote));
   g_assert_cmpint (flatpak_remote_get_prio (remote), ==, 1);
 
-#ifdef FLATPAK_ENABLE_P2P
   g_assert_cmpstr (flatpak_remote_get_collection_id (remote), ==, repo_collection_id);
-#endif  /* FLATPAK_ENABLE_P2P */
 }
 
 static void
@@ -273,6 +304,10 @@ test_remote (void)
   g_autoptr(FlatpakInstallation) inst = NULL;
   g_autoptr(GError) error = NULL;
   g_autoptr(FlatpakRemote) remote = NULL;
+  g_autoptr(GFile) inst_file = NULL;
+  g_autoptr(GFile) repo_file = NULL;
+  g_autoptr(OstreeRepo) repo = NULL;
+  gboolean gpg_verify_summary;
   gboolean res;
 
   inst = flatpak_installation_new_user (NULL, &error);
@@ -281,14 +316,40 @@ test_remote (void)
   remote = flatpak_installation_get_remote_by_name (inst, repo_name, NULL, &error);
   g_assert_no_error (error);
 
-#ifdef FLATPAK_ENABLE_P2P
-  g_assert_cmpstr (flatpak_remote_get_collection_id (remote), ==, NULL);
-  flatpak_remote_set_collection_id (remote, "org.example.CollectionID");
-  g_assert_cmpstr (flatpak_remote_get_collection_id (remote), ==, "org.example.CollectionID");
-  /* Don’t leave the collection ID set since the repos aren’t configured with one. */
+  g_assert_cmpstr (flatpak_remote_get_collection_id (remote), ==, repo_collection_id);
+
+  /* Flatpak doesn't provide access to gpg-verify-summary, so use ostree */
+  res = flatpak_installation_modify_remote (inst, remote, NULL, &error);
+  g_assert_no_error (error);
+  g_assert_true (res);
+  inst_file = flatpak_installation_get_path (inst);
+  repo_file = g_file_get_child (inst_file, "repo");
+  repo = ostree_repo_new (repo_file);
+  res = ostree_repo_open (repo, NULL, &error);
+  g_assert_no_error (error);
+  g_assert_true (res);
+  res = ostree_repo_get_remote_boolean_option (repo, repo_name, "gpg-verify-summary", TRUE, &gpg_verify_summary, &error);
+  g_assert_no_error (error);
+  g_assert_true (res);
+  g_assert_false (gpg_verify_summary);
+
+  /* Temporarily unset the collection ID */
   flatpak_remote_set_collection_id (remote, NULL);
   g_assert_cmpstr (flatpak_remote_get_collection_id (remote), ==, NULL);
-#endif  /* FLATPAK_ENABLE_P2P */
+
+  res = flatpak_installation_modify_remote (inst, remote, NULL, &error);
+  g_assert_no_error (error);
+  g_assert_true (res);
+  res = ostree_repo_reload_config (repo, NULL, &error);
+  g_assert_no_error (error);
+  g_assert_true (res);
+  res = ostree_repo_get_remote_boolean_option (repo, repo_name, "gpg-verify-summary", FALSE, &gpg_verify_summary, &error);
+  g_assert_no_error (error);
+  g_assert_true (res);
+  g_assert_true (gpg_verify_summary);
+
+  flatpak_remote_set_collection_id (remote, repo_collection_id);
+  g_assert_cmpstr (flatpak_remote_get_collection_id (remote), ==, repo_collection_id);
 
   g_assert_cmpstr (flatpak_remote_get_title (remote), ==, NULL);
   flatpak_remote_set_title (remote, "Test Repo");
@@ -299,8 +360,8 @@ test_remote (void)
   g_assert_cmpint (flatpak_remote_get_prio (remote), ==, 15);
 
   res = flatpak_installation_modify_remote (inst, remote, NULL, &error);
-  g_assert_true (res);
   g_assert_no_error (error);
+  g_assert_true (res);
 
   g_clear_object (&remote);
 
@@ -327,24 +388,168 @@ test_list_refs (void)
   g_assert_cmpint (refs->len, ==, 0);
 }
 
+static void
+create_multi_collection_id_repo (const char *repo_dir)
+{
+  g_autoptr(GError) error = NULL;
+  g_autofree char *arg0 = NULL;
+
+  /* Create a repository in which each app has a different collection-id */
+  arg0 = g_test_build_filename (G_TEST_DIST, "make-multi-collection-id-repo.sh", NULL);
+  const char *argv[] = { arg0, repo_dir, NULL };
+  run_test_subprocess ((char **) argv, RUN_TEST_SUBPROCESS_DEFAULT);
+}
+
+static void
+test_list_refs_in_remotes (void)
+{
+  const char *repo_name = "multi-refs-repo";
+  g_autofree char *repo_url = NULL;
+
+  g_autoptr(GPtrArray) refs1 = NULL;
+  g_autoptr(GPtrArray) refs2 = NULL;
+  g_autoptr(FlatpakInstallation) inst = NULL;
+  g_autoptr(GError) error = NULL;
+  g_autoptr(FlatpakRemote) remote = NULL;
+  g_autofree char *repo_dir = g_build_filename (testdir, repo_name, NULL);
+  g_autofree char *repo_uri = NULL;
+  g_autoptr(GHashTable) collection_ids = g_hash_table_new_full (g_str_hash,
+                                                                g_str_equal,
+                                                                NULL, NULL);
+  g_autoptr(GHashTable) ref_specs = g_hash_table_new_full (g_str_hash,
+                                                           g_str_equal,
+                                                           g_free,
+                                                           NULL);
+
+  create_multi_collection_id_repo (repo_dir);
+
+  repo_url = g_strdup_printf ("file://%s", repo_dir);
+
+  const char *argv[] = { "flatpak", "remote-add", "--user", "--no-gpg-verify",
+                         repo_name, repo_url, NULL };
+
+  /* Add the repo we created above, which holds one collection ID per ref */
+  run_test_subprocess ((char **) argv, RUN_TEST_SUBPROCESS_DEFAULT);
+
+  inst = flatpak_installation_new_user (NULL, &error);
+  g_assert_no_error (error);
+
+  /* Ensure the remote can be successfully found */
+  remote = flatpak_installation_get_remote_by_name (inst, repo_name, NULL, &error);
+  g_assert_no_error (error);
+  g_assert_nonnull (remote);
+
+  /* List the refs in the remote we've just added */
+  refs1 = flatpak_installation_list_remote_refs_sync (inst, repo_name, NULL, &error);
+
+  g_assert_no_error (error);
+  g_assert_nonnull (refs1);
+  g_assert (refs1->len > 1);
+
+  /* Ensure that the number of different collection IDs is the same as the
+   * number of apps */
+  for (guint i = 0; i < refs1->len; ++i)
+    {
+      FlatpakRef *ref = g_ptr_array_index (refs1, i);
+      g_hash_table_add (collection_ids, (gchar *) flatpak_ref_get_collection_id (ref));
+      g_hash_table_add (ref_specs, flatpak_ref_format_ref (ref));
+    }
+
+  g_assert_cmpuint (g_hash_table_size (collection_ids), ==, refs1->len);
+
+  /* Ensure that listing the refs by using a remote's URI will get us the
+   * same results as using the name */
+  repo_uri = flatpak_remote_get_url (remote);
+  refs2 = flatpak_installation_list_remote_refs_sync (inst, repo_uri, NULL, &error);
+
+  g_assert_no_error (error);
+  g_assert_nonnull (refs2);
+  g_assert_cmpuint (refs2->len, ==, refs1->len);
+
+  for (guint i = 0; i < refs2->len; ++i)
+    {
+      FlatpakRef *ref = g_ptr_array_index (refs2, i);
+      g_autofree char *ref_spec = flatpak_ref_format_ref (ref);
+      g_assert_nonnull (g_hash_table_lookup (ref_specs, ref_spec));
+    }
+}
+
+static void
+test_list_remote_refs (void)
+{
+  g_autoptr(FlatpakInstallation) inst = NULL;
+  g_autoptr(GError) error = NULL;
+  g_autoptr(GPtrArray) refs = NULL;
+  int i;
+
+  inst = flatpak_installation_new_user (NULL, &error);
+  g_assert_no_error (error);
+
+  refs = flatpak_installation_list_remote_refs_sync (inst, repo_name, NULL, &error);
+  g_assert_no_error (error);
+  g_assert_nonnull (refs);
+  g_assert_cmpint (refs->len, ==, 2);
+
+
+  for (i = 0; i < refs->len; i++)
+    {
+      FlatpakRemoteRef *remote_ref = g_ptr_array_index (refs, i);
+      FlatpakRef *ref = FLATPAK_REF (remote_ref);
+      GBytes *metadata;
+
+      g_assert (ref != NULL);
+
+      if (strcmp ("org.test.Hello", flatpak_ref_get_name (ref)) == 0)
+        {
+          g_assert_cmpint (flatpak_ref_get_kind (ref), ==, FLATPAK_REF_KIND_APP);
+        }
+      else
+        {
+          g_assert_cmpstr (flatpak_ref_get_name (ref), ==, "org.test.Platform");
+          g_assert_cmpint (flatpak_ref_get_kind (ref), ==, FLATPAK_REF_KIND_RUNTIME);
+        }
+
+      g_assert_cmpstr (flatpak_ref_get_branch (ref), ==, "master");
+      g_assert_cmpstr (flatpak_ref_get_commit (ref), !=, NULL);
+      g_assert_cmpstr (flatpak_ref_get_arch (ref), ==, flatpak_get_default_arch ());
+
+      g_assert_cmpstr (flatpak_remote_ref_get_remote_name (remote_ref), ==, repo_name);
+      g_assert_cmpstr (flatpak_remote_ref_get_eol (remote_ref), ==, NULL);
+      g_assert_cmpstr (flatpak_remote_ref_get_eol_rebase (remote_ref), ==, NULL);
+
+      g_assert_cmpuint (flatpak_remote_ref_get_installed_size (remote_ref), >, 0);
+      g_assert_cmpuint (flatpak_remote_ref_get_download_size (remote_ref), >, 0);
+
+      metadata = flatpak_remote_ref_get_metadata (remote_ref);
+      g_assert (metadata != NULL);
+
+      if (strcmp ("org.test.Hello", flatpak_ref_get_name (ref)) == 0)
+        g_assert (g_str_has_prefix ((char *) g_bytes_get_data (metadata, NULL), "[Application]"));
+      else
+        g_assert (g_str_has_prefix ((char *) g_bytes_get_data (metadata, NULL), "[Runtime]"));
+    }
+}
+
 static void
 progress_cb (const char *status,
-             guint progress,
-             gboolean estimating,
-             gpointer user_data)
+             guint       progress,
+             gboolean    estimating,
+             gpointer    user_data)
 {
   int *count = user_data;
+
   *count += 1;
 }
 
 static void
-changed_cb (GFileMonitor *monitor,
-            GFile *file,
-            GFile *other_file,
+changed_cb (GFileMonitor     *monitor,
+            GFile            *file,
+            GFile            *other_file,
             GFileMonitorEvent event_type,
-            gpointer user_data)
+            gpointer          user_data)
 {
   int *count = user_data;
+
   *count += 1;
 }
 
@@ -352,6 +557,7 @@ static gboolean
 timeout_cb (gpointer data)
 {
   gboolean *timeout_reached = data;
+
   *timeout_reached = TRUE;
   return G_SOURCE_CONTINUE;
 }
@@ -377,8 +583,8 @@ test_install_launch_uninstall (void)
   if (bwrap != NULL)
     {
       gint exit_code = 0;
-      char *argv[] = { (char *)bwrap, "--unshare-ipc", "--unshare-net",
-          "--unshare-pid", "--ro-bind", "/", "/", "/bin/true", NULL };
+      char *argv[] = { (char *) bwrap, "--unshare-ipc", "--unshare-net",
+                       "--unshare-pid", "--ro-bind", "/", "/", "/bin/true", NULL };
       g_autofree char *argv_str = g_strjoinv (" ", argv);
       g_test_message ("Spawning %s", argv_str);
       g_spawn_sync (NULL, argv, NULL, G_SPAWN_SEARCH_PATH, NULL, NULL, NULL, NULL, &exit_code, &error);
@@ -433,9 +639,7 @@ test_install_launch_uninstall (void)
   g_assert_cmpstr (flatpak_ref_get_arch (FLATPAK_REF (ref)), ==, flatpak_get_default_arch ());
   g_assert_cmpstr (flatpak_ref_get_branch (FLATPAK_REF (ref)), ==, "master");
   g_assert_cmpint (flatpak_ref_get_kind (FLATPAK_REF (ref)), ==, FLATPAK_REF_KIND_RUNTIME);
-#ifdef FLATPAK_ENABLE_P2P
   g_assert_null (flatpak_ref_get_collection_id (FLATPAK_REF (ref)));
-#endif  /* FLATPAK_ENABLE_P2P */
 
   g_assert_cmpuint (flatpak_installed_ref_get_installed_size (ref), >, 0);
 
@@ -483,9 +687,7 @@ test_install_launch_uninstall (void)
   g_assert_cmpstr (flatpak_ref_get_arch (FLATPAK_REF (ref)), ==, flatpak_get_default_arch ());
   g_assert_cmpstr (flatpak_ref_get_branch (FLATPAK_REF (ref)), ==, "master");
   g_assert_cmpint (flatpak_ref_get_kind (FLATPAK_REF (ref)), ==, FLATPAK_REF_KIND_APP);
-#ifdef FLATPAK_ENABLE_P2P
   g_assert_null (flatpak_ref_get_collection_id (FLATPAK_REF (ref)));
-#endif  /* FLATPAK_ENABLE_P2P */
 
   g_assert_cmpuint (flatpak_installed_ref_get_installed_size (ref), >, 0);
   g_assert_true (flatpak_installed_ref_get_is_current (ref));
@@ -563,18 +765,86 @@ test_install_launch_uninstall (void)
   g_ptr_array_unref (refs);
 }
 
-typedef enum
-{
-  RUN_TEST_SUBPROCESS_DEFAULT = 0,
-  RUN_TEST_SUBPROCESS_IGNORE_FAILURE = (1 << 0),
-  RUN_TEST_SUBPROCESS_NO_CAPTURE = (1 << 1),
-} RunTestSubprocessFlags;
+static void update_test_app (void);
+static void update_repo (void);
 
 static void
-run_test_subprocess (char **argv,
+test_list_updates (void)
+{
+  g_autoptr(FlatpakInstallation) inst = NULL;
+  g_autoptr(GError) error = NULL;
+  g_autoptr(GPtrArray) refs = NULL;
+  g_autoptr(FlatpakInstalledRef) ref = NULL;
+  g_autoptr(FlatpakInstalledRef) runtime_ref = NULL;
+  FlatpakInstalledRef *update_ref = NULL;
+  gboolean res;
+
+  inst = flatpak_installation_new_user (NULL, &error);
+  g_assert_no_error (error);
+
+  /* Install a runtime and app */
+  runtime_ref = flatpak_installation_install (inst,
+                                              repo_name,
+                                              FLATPAK_REF_KIND_RUNTIME,
+                                              "org.test.Platform",
+                                              NULL, NULL, NULL, NULL, NULL,
+                                              &error);
+  g_assert_no_error (error);
+  g_assert (FLATPAK_IS_INSTALLED_REF (runtime_ref));
+
+  ref = flatpak_installation_install (inst,
+                                      repo_name,
+                                      FLATPAK_REF_KIND_APP,
+                                      "org.test.Hello",
+                                      NULL, NULL, NULL, NULL, NULL,
+                                      &error);
+  g_assert_no_error (error);
+  g_assert (FLATPAK_IS_INSTALLED_REF (ref));
+
+  /* Update the test app and list the update */
+  update_test_app ();
+  update_repo ();
+
+  /* Drop all in-memory summary caches so we can find the new update */
+  flatpak_installation_drop_caches (inst, NULL, &error);
+  g_assert_no_error (error);
+
+  refs = flatpak_installation_list_installed_refs_for_update (inst, NULL, &error);
+  g_assert_no_error (error);
+  g_assert_nonnull (refs);
+  g_assert_cmpint (refs->len, ==, 1);
+  update_ref = g_ptr_array_index (refs, 0);
+  g_assert_cmpstr (flatpak_ref_get_name (FLATPAK_REF (update_ref)), ==, "org.test.Hello");
+  g_assert_cmpint (flatpak_ref_get_kind (FLATPAK_REF (update_ref)), ==, FLATPAK_REF_KIND_APP);
+
+  /* Uninstall the runtime and app */
+  res = flatpak_installation_uninstall (inst,
+                                        flatpak_ref_get_kind (FLATPAK_REF (ref)),
+                                        flatpak_ref_get_name (FLATPAK_REF (ref)),
+                                        flatpak_ref_get_arch (FLATPAK_REF (ref)),
+                                        flatpak_ref_get_branch (FLATPAK_REF (ref)),
+                                        NULL, NULL, NULL,
+                                        &error);
+  g_assert_no_error (error);
+  g_assert_true (res);
+
+  res = flatpak_installation_uninstall (inst,
+                                        flatpak_ref_get_kind (FLATPAK_REF (runtime_ref)),
+                                        flatpak_ref_get_name (FLATPAK_REF (runtime_ref)),
+                                        flatpak_ref_get_arch (FLATPAK_REF (runtime_ref)),
+                                        flatpak_ref_get_branch (FLATPAK_REF (runtime_ref)),
+                                        NULL, NULL, NULL,
+                                        &error);
+  g_assert_no_error (error);
+  g_assert_true (res);
+}
+
+static void
+run_test_subprocess (char                 **argv,
                      RunTestSubprocessFlags flags)
 {
   int status;
+
   g_autoptr(GError) error = NULL;
   g_autofree char *argv_str = g_strjoinv (" ", argv);
   g_autofree char *output = NULL;
@@ -583,7 +853,7 @@ run_test_subprocess (char **argv,
   g_test_message ("Spawning %s", argv_str);
 
   if (flags & RUN_TEST_SUBPROCESS_NO_CAPTURE)
-    g_spawn_sync (NULL, argv, NULL, G_SPAWN_SEARCH_PATH | G_SPAWN_STDOUT_TO_DEV_NULL | G_SPAWN_STDERR_TO_DEV_NULL, NULL, NULL, NULL, NULL, &status, &error);
+    g_spawn_sync (NULL, argv, NULL, G_SPAWN_SEARCH_PATH | G_SPAWN_STDOUT_TO_DEV_NULL, NULL, NULL, NULL, NULL, &status, &error);
   else
     g_spawn_sync (NULL, argv, NULL, G_SPAWN_SEARCH_PATH, NULL, NULL, &output, &errors, &status, &error);
 
@@ -629,6 +899,8 @@ make_test_runtime (void)
 
   arg0 = g_test_build_filename (G_TEST_DIST, "make-test-runtime.sh", NULL);
   argv[0] = arg0;
+  argv[3] = repo_collection_id;
+
   run_test_subprocess (argv, RUN_TEST_SUBPROCESS_DEFAULT);
 }
 
@@ -636,10 +908,25 @@ static void
 make_test_app (void)
 {
   g_autofree char *arg0 = NULL;
-  char *argv[] = { NULL, "test", "", NULL };
+  char *argv[] = { NULL, "test", "", "", NULL };
 
   arg0 = g_test_build_filename (G_TEST_DIST, "make-test-app.sh", NULL);
   argv[0] = arg0;
+  argv[3] = repo_collection_id;
+
+  run_test_subprocess (argv, RUN_TEST_SUBPROCESS_DEFAULT);
+}
+
+static void
+update_test_app (void)
+{
+  g_autofree char *arg0 = NULL;
+  char *argv[] = { NULL, "test", "", "", "UPDATED", NULL };
+
+  arg0 = g_test_build_filename (G_TEST_DIST, "make-test-app.sh", NULL);
+  argv[0] = arg0;
+  argv[3] = repo_collection_id;
+
   run_test_subprocess (argv, RUN_TEST_SUBPROCESS_DEFAULT);
 }
 
@@ -647,6 +934,7 @@ static void
 update_repo (void)
 {
   char *argv[] = { "flatpak", "build-update-repo", "--gpg-homedir=", "--gpg-sign=", "repos/test", NULL };
+
   g_auto(GStrv) gpgargs = NULL;
 
   gpgargs = g_strsplit (gpg_args, " ", 0);
@@ -659,7 +947,7 @@ static void
 launch_httpd (void)
 {
   g_autofree char *path = g_test_build_filename (G_TEST_DIST, "test-webserver.sh", NULL);
-  char *argv[] = {path , "repos", NULL };
+  char *argv[] = {path, "repos", NULL };
 
   /* The web server puts itself in the background, so we can't wait
    * for EOF on its stdout, stderr */
@@ -670,11 +958,11 @@ static void
 add_remote (void)
 {
   g_autoptr(GError) error = NULL;
-  char *argv[] = { "flatpak", "remote-add", "--user", "--gpg-import=", "name", "url", NULL };
-  g_autofree char *argv_str = NULL;
+  char *argv[] = { "flatpak", "remote-add", "--user", "--gpg-import=", "--collection-id=", "name", "url", NULL };
   g_autofree char *gpgimport = NULL;
   g_autofree char *port = NULL;
   g_autofree char *pid = NULL;
+  g_autofree char *collection_id_arg = NULL;
 
   launch_httpd ();
 
@@ -692,10 +980,12 @@ add_remote (void)
 
   gpgimport = g_strdup_printf ("--gpg-import=%s/pubring.gpg", gpg_homedir);
   repo_url = g_strdup_printf ("http://127.0.0.1:%s/test", port);
+  collection_id_arg = g_strdup_printf ("--collection-id=%s", repo_collection_id);
 
   argv[3] = gpgimport;
-  argv[4] = (char *)repo_name;
-  argv[5] = repo_url;
+  argv[4] = collection_id_arg;
+  argv[5] = (char *) repo_name;
+  argv[6] = repo_url;
   run_test_subprocess (argv, RUN_TEST_SUBPROCESS_DEFAULT);
 }
 
@@ -708,18 +998,19 @@ add_extra_installation (const char *id,
   g_autofree char *conffile_path = NULL;
   g_autofree char *contents_string = NULL;
   g_autofree char *path = NULL;
+
   g_autoptr(GPtrArray) contents_array = NULL;
   g_autoptr(GError) error = NULL;
 
   path = g_strconcat (testdir, "/system-", id, NULL);
-  g_mkdir_with_parents (path, S_IRWXU|S_IRWXG|S_IRWXO);
+  g_mkdir_with_parents (path, S_IRWXU | S_IRWXG | S_IRWXO);
 
-  contents_array = g_ptr_array_new_with_free_func ((GDestroyNotify)g_free);
+  contents_array = g_ptr_array_new_with_free_func ((GDestroyNotify) g_free);
 
   g_ptr_array_add (contents_array,
-      g_strdup_printf ("[Installation \"%s\"]\n"
-                       "Path=%s",
-                       id, path));
+                   g_strdup_printf ("[Installation \"%s\"]\n"
+                                    "Path=%s",
+                                    id, path));
 
   if (display_name != NULL)
     g_ptr_array_add (contents_array, g_strdup_printf ("DisplayName=%s", display_name));
@@ -731,7 +1022,7 @@ add_extra_installation (const char *id,
     g_ptr_array_add (contents_array, g_strdup_printf ("Priority=%s", priority));
 
   g_ptr_array_add (contents_array, NULL);
-  contents_string = g_strjoinv ("\n", (char**)contents_array->pdata);
+  contents_string = g_strjoinv ("\n", (char **) contents_array->pdata);
 
   conffile_path = g_strconcat (flatpak_installationsdir, "/", id, ".conf", NULL);
   g_file_set_contents (conffile_path, contents_string, -1, &error);
@@ -742,7 +1033,7 @@ static void
 setup_multiple_installations (void)
 {
   flatpak_installationsdir = g_strconcat (flatpak_configdir, "/installations.d", NULL);
-  g_mkdir_with_parents (flatpak_installationsdir, S_IRWXU|S_IRWXG|S_IRWXO);
+  g_mkdir_with_parents (flatpak_installationsdir, S_IRWXU | S_IRWXG | S_IRWXO);
 
   add_extra_installation ("extra-installation-1", "Extra system installation 1", "mmc", "10");
   add_extra_installation ("extra-installation-2", "Extra system installation 2", "sdcard", "25");
@@ -752,6 +1043,8 @@ setup_multiple_installations (void)
 static void
 setup_repo (void)
 {
+  repo_collection_id = "com.example.Test";
+
   make_test_runtime ();
   make_test_app ();
   update_repo ();
@@ -763,6 +1056,7 @@ copy_file (const char *src, const char *dest)
 {
   gchar *buffer = NULL;
   gsize length;
+
   g_autoptr(GError) error = NULL;
 
   g_test_message ("copying %s to %s", src, dest);
@@ -805,54 +1099,54 @@ global_setup (void)
   g_test_message ("testdir: %s", testdir);
 
   homedir = g_strconcat (testdir, "/home", NULL);
-  g_mkdir_with_parents (homedir, S_IRWXU|S_IRWXG|S_IRWXO);
+  g_mkdir_with_parents (homedir, S_IRWXU | S_IRWXG | S_IRWXO);
 
   g_setenv ("HOME", homedir, TRUE);
-  g_test_message ("setting HOME=%s", datadir);
+  g_test_message ("setting HOME=%s", homedir);
 
   cachedir = g_strconcat (testdir, "/home/cache", NULL);
-  g_mkdir_with_parents (cachedir, S_IRWXU|S_IRWXG|S_IRWXO);
+  g_mkdir_with_parents (cachedir, S_IRWXU | S_IRWXG | S_IRWXO);
   g_setenv ("XDG_CACHE_HOME", cachedir, TRUE);
   g_test_message ("setting XDG_CACHE_HOME=%s", cachedir);
 
   configdir = g_strconcat (testdir, "/home/config", NULL);
-  g_mkdir_with_parents (configdir, S_IRWXU|S_IRWXG|S_IRWXO);
+  g_mkdir_with_parents (configdir, S_IRWXU | S_IRWXG | S_IRWXO);
   g_setenv ("XDG_CONFIG_HOME", configdir, TRUE);
   g_test_message ("setting XDG_CONFIG_HOME=%s", configdir);
 
   datadir = g_strconcat (testdir, "/home/share", NULL);
-  g_mkdir_with_parents (datadir, S_IRWXU|S_IRWXG|S_IRWXO);
+  g_mkdir_with_parents (datadir, S_IRWXU | S_IRWXG | S_IRWXO);
   g_setenv ("XDG_DATA_HOME", datadir, TRUE);
   g_test_message ("setting XDG_DATA_HOME=%s", datadir);
 
   flatpak_runtimedir = g_strconcat (testdir, "/runtime", NULL);
-  g_mkdir_with_parents (flatpak_runtimedir, S_IRWXU|S_IRWXG|S_IRWXO);
+  g_mkdir_with_parents (flatpak_runtimedir, S_IRWXU | S_IRWXG | S_IRWXO);
   g_test_message ("setting XDG_RUNTIME_DIR=%s", flatpak_runtimedir);
 
   flatpak_systemdir = g_strconcat (testdir, "/system", NULL);
-  g_mkdir_with_parents (flatpak_systemdir, S_IRWXU|S_IRWXG|S_IRWXO);
+  g_mkdir_with_parents (flatpak_systemdir, S_IRWXU | S_IRWXG | S_IRWXO);
   g_setenv ("FLATPAK_SYSTEM_DIR", flatpak_systemdir, TRUE);
   g_test_message ("setting FLATPAK_SYSTEM_DIR=%s", flatpak_systemdir);
 
   flatpak_systemcachedir = g_strconcat (testdir, "/system-cache", NULL);
-  g_mkdir_with_parents (flatpak_systemcachedir, S_IRWXU|S_IRWXG|S_IRWXO);
+  g_mkdir_with_parents (flatpak_systemcachedir, S_IRWXU | S_IRWXG | S_IRWXO);
   g_setenv ("FLATPAK_SYSTEM_CACHE_DIR", flatpak_systemcachedir, TRUE);
   g_test_message ("setting FLATPAK_SYSTEM_CACHE_DIR=%s", flatpak_systemcachedir);
 
   flatpak_configdir = g_strconcat (testdir, "/config", NULL);
-  g_mkdir_with_parents (flatpak_configdir, S_IRWXU|S_IRWXG|S_IRWXO);
+  g_mkdir_with_parents (flatpak_configdir, S_IRWXU | S_IRWXG | S_IRWXO);
   g_setenv ("FLATPAK_CONFIG_DIR", flatpak_configdir, TRUE);
   g_test_message ("setting FLATPAK_CONFIG_DIR=%s", flatpak_configdir);
 
   gpg_homedir = g_strconcat (testdir, "/gpghome", NULL);
-  g_mkdir_with_parents (gpg_homedir, S_IRWXU|S_IRWXG|S_IRWXO);
+  g_mkdir_with_parents (gpg_homedir, S_IRWXU | S_IRWXG | S_IRWXO);
 
   gpg_args = g_strdup_printf ("--gpg-homedir=%s --gpg-sign=%s", gpg_homedir, gpg_id);
   g_setenv ("GPGARGS", gpg_args, TRUE);
   g_test_message ("setting GPGARGS=%s", gpg_args);
 
   copy_gpg ();
-  setup_multiple_installations();
+  setup_multiple_installations ();
   setup_repo ();
 }
 
@@ -891,8 +1185,11 @@ main (int argc, char *argv[])
   g_test_add_func ("/library/list-remotes", test_list_remotes);
   g_test_add_func ("/library/remote-by-name", test_remote_by_name);
   g_test_add_func ("/library/remote", test_remote);
+  g_test_add_func ("/library/list-remote-refs", test_list_remote_refs);
   g_test_add_func ("/library/list-refs", test_list_refs);
   g_test_add_func ("/library/install-launch-uninstall", test_install_launch_uninstall);
+  g_test_add_func ("/library/list-refs-in-remote", test_list_refs_in_remotes);
+  g_test_add_func ("/library/list-updates", test_list_updates);
 
   global_setup ();
 
